@@ -3,7 +3,7 @@ from unittest.mock import Mock, patch
 
 from django.contrib.auth import get_user_model
 from django.db import IntegrityError, transaction
-from django.test import TestCase, TransactionTestCase
+from django.test import TestCase, TransactionTestCase, override_settings
 from django.utils import timezone
 
 from simple_history.exceptions import AlternativeManagerError, NotHistoricalModelError
@@ -77,6 +77,13 @@ class BulkCreateWithHistoryTestCase(TestCase):
         self.assertEqual(Poll.objects.count(), 5)
         self.assertEqual(Poll.history.count(), 5)
 
+    @override_settings(SIMPLE_HISTORY_ENABLED=False)
+    def test_bulk_create_history_with_disabled_setting(self):
+        bulk_create_with_history(self.data, Poll)
+
+        self.assertEqual(Poll.objects.count(), 5)
+        self.assertEqual(Poll.history.count(), 0)
+
     def test_bulk_create_history_alternative_manager(self):
         bulk_create_with_history(
             self.data,
@@ -179,6 +186,38 @@ class BulkCreateWithHistoryTestCase(TestCase):
         self.assertEqual(PollWithUniqueQuestion.objects.count(), 2)
         self.assertEqual(PollWithUniqueQuestion.history.count(), 2)
 
+    def test_bulk_create_history_with_no_ids_return(self):
+        pub_date = timezone.now()
+        objects = [
+            Poll(question="Question 1", pub_date=pub_date),
+            Poll(question="Question 2", pub_date=pub_date),
+            Poll(question="Question 3", pub_date=pub_date),
+            Poll(question="Question 4", pub_date=pub_date),
+            Poll(question="Question 5", pub_date=pub_date),
+        ]
+
+        _bulk_create = Poll._default_manager.bulk_create
+
+        def mock_bulk_create(*args, **kwargs):
+            _bulk_create(*args, **kwargs)
+            return [
+                Poll(question="Question 1", pub_date=pub_date),
+                Poll(question="Question 2", pub_date=pub_date),
+                Poll(question="Question 3", pub_date=pub_date),
+                Poll(question="Question 4", pub_date=pub_date),
+                Poll(question="Question 5", pub_date=pub_date),
+            ]
+
+        with patch.object(
+            Poll._default_manager, "bulk_create", side_effect=mock_bulk_create
+        ):
+            with self.assertNumQueries(3):
+                result = bulk_create_with_history(objects, Poll)
+            self.assertEqual(
+                [poll.question for poll in result], [poll.question for poll in objects]
+            )
+            self.assertNotEqual(result[0].id, None)
+
 
 class BulkCreateWithHistoryTransactionTestCase(TransactionTestCase):
     def setUp(self):
@@ -235,7 +274,7 @@ class BulkCreateWithHistoryTransactionTestCase(TransactionTestCase):
         model = Mock(
             _default_manager=Mock(
                 bulk_create=Mock(return_value=[Place(name="Place 1")]),
-                filter=Mock(return_value=objects),
+                filter=Mock(return_value=Mock(order_by=Mock(return_value=objects))),
             ),
             _meta=Mock(get_fields=Mock(return_value=[])),
         )
@@ -291,6 +330,21 @@ class BulkUpdateWithHistoryTestCase(TestCase):
         self.assertEqual(Poll.history.count(), 10)
         self.assertEqual(Poll.history.filter(history_type="~").count(), 5)
 
+    @override_settings(SIMPLE_HISTORY_ENABLED=False)
+    def test_bulk_update_history_without_history_enabled(self):
+        self.assertEqual(Poll.history.count(), 5)
+        # because setup called with enabled settings
+        bulk_update_with_history(
+            self.data,
+            Poll,
+            fields=["question"],
+        )
+
+        self.assertEqual(Poll.objects.count(), 5)
+        self.assertEqual(Poll.objects.get(id=4).question, "Updated question")
+        self.assertEqual(Poll.history.count(), 5)
+        self.assertEqual(Poll.history.filter(history_type="~").count(), 0)
+
     def test_bulk_update_history_with_default_user(self):
         user = User.objects.create_user("tester", "tester@example.com")
 
